@@ -20,6 +20,7 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import numpy as np
 from scipy.ndimage import generic_filter
+import matplotlib.pyplot as plt
 
 
 def global_val(u, v, u_thresholds, v_thresholds):
@@ -71,13 +72,13 @@ def global_val(u, v, u_thresholds, v_thresholds):
     u[ind] = np.nan
     v[ind] = np.nan
 
-    mask = np.zeros(u.shape, dtype=bool)
+    mask = np.zeros_like(u, dtype=bool)
     mask[ind] = True
 
     return u, v, mask
 
 
-def global_std(u, v, std_threshold=3):
+def global_std(u, v, std_threshold=5):
     """Eliminate spurious vectors with a global threshold defined by the
     standard deviation
 
@@ -115,23 +116,31 @@ def global_std(u, v, std_threshold=3):
     # both previous nans and masked regions are not 
     # participating in the magnitude comparison
 
-    if isinstance(u, np.ma.MaskedArray):
-        vel_magnitude = u.data ** 2 + v.data ** 2
-    else:
-        vel_magnitude = u ** 2 + v ** 2
+    # def reject_outliers(data, m=2):
+    #     return data[abs(data - np.mean(data)) < m * np.std(data)]    
 
-    ind = vel_magnitude > std_threshold * np.std(vel_magnitude)
+    # create nan filled arrays where masks
+    # if u,v, are non-masked, ma.copy() adds false masks
+    tmpu = np.ma.copy(u).filled(np.nan)
+    tmpv = np.ma.copy(v).filled(np.nan)
+
+    ind = np.logical_or(np.abs(tmpu - np.nanmean(tmpu)) > std_threshold * np.nanstd(tmpu), 
+                        np.abs(tmpv - np.nanmean(tmpv)) > std_threshold * np.nanstd(tmpv))
+
+    if np.all(ind): # if all is True, something is really wrong
+        print('Warning! probably a uniform shift data, do not use this filter')
+        ind = ~ind
 
     u[ind] = np.nan
     v[ind] = np.nan
 
-    mask = np.zeros(u.shape, dtype=bool)
+    mask = np.zeros_like(u, dtype=bool)
     mask[ind] = True
 
     return u, v, mask
 
 
-def sig2noise_val(u, v, sig2noise, w=None, threshold=1.05):
+def sig2noise_val(u, v, s2n, w=None, threshold=1.05):
     """Eliminate spurious vectors from cross-correlation signal to noise ratio.
 
     Replace spurious vectors with zero if signal to noise ratio
@@ -145,7 +154,7 @@ def sig2noise_val(u, v, sig2noise, w=None, threshold=1.05):
     v : 2d or 3d np.ndarray
         a two or three dimensional array containing the v velocity component.
 
-    sig2noise : 2d np.ndarray
+    s2n : 2d np.ndarray
         a two or three dimensional array containing the value  of the signal to
         noise ratio from cross-correlation function.
     w : 2d or 3d np.ndarray
@@ -179,13 +188,12 @@ def sig2noise_val(u, v, sig2noise, w=None, threshold=1.05):
         1, 1202-1215.
 
     """
-
-    ind = sig2noise < threshold
+    ind = s2n < threshold
 
     u[ind] = np.nan
     v[ind] = np.nan
 
-    mask = np.zeros(u.shape, dtype=bool)
+    mask = np.zeros_like(u, dtype=bool)
     mask[ind] = True
 
     if isinstance(w, np.ndarray):
@@ -240,11 +248,12 @@ def local_median_val(u, v, u_threshold, v_threshold, size=1):
     # NaN and then use generic filter with nanmean
     # 
 
-    u = np.ma.MaskedArray(u)
-    v = np.ma.MaskedArray(v)
+    u = np.ma.copy(u)
+    v = np.ma.copy(v)
     
     # kernel footprint
     f = np.ones((2*size+1, 2*size+1))
+    f[size,size] = 0
 
     masked_u = np.where(~u.mask, u.data, np.nan)
     masked_v = np.where(~v.mask, v.data, np.nan)
@@ -263,3 +272,83 @@ def local_median_val(u, v, u_threshold, v_threshold, size=1):
     mask[ind] = True
 
     return u, v, mask
+
+
+def typical_validation(u, v, s2n, settings):
+
+    """
+    validation using gloabl limits and std and local median, 
+
+    with a special option of 'no_std' for the case of completely
+    uniform shift, e.g. in tests. 
+
+    see Settings() for the parameters:
+
+    MinMaxU : two elements tuple
+        sets the limits of the u displacment component
+        Used for validation.
+
+    MinMaxV : two elements tuple
+        sets the limits of the v displacment component
+        Used for validation.
+
+    std_threshold : float
+        sets the  threshold for the std validation
+
+    median_threshold : float
+        sets the threshold for the median validation
+    """
+
+    if settings.show_all_plots:
+        plt.figure()
+        plt.quiver(u,v,color='b')
+
+    mask = np.zeros(u.shape, dtype=bool)
+
+    u, v, mask_g = global_val(
+        u, v, settings.MinMax_U_disp, settings.MinMax_V_disp
+    )
+    # print(f"global filter invalidated {sum(mask_g.flatten())} vectors")
+    if settings.show_all_plots:
+        plt.quiver(u,v,color='m')
+
+    u, v, mask_s = global_std(
+        u, v, std_threshold=settings.std_threshold
+    )
+    # print(f"std filter invalidated {sum(mask_s.flatten())} vectors")
+    if settings.show_all_plots:
+        plt.quiver(u,v,color='k')
+    
+
+    u, v, mask_m = local_median_val(
+        u,
+        v,
+        u_threshold=settings.median_threshold,
+        v_threshold=settings.median_threshold,
+        size=settings.median_size,
+    )
+    if settings.show_all_plots:
+        plt.quiver(u,v,color='r')
+
+    # print(f"median filter invalidated {sum(mask_m.flatten())} vectors")
+    mask = mask + mask_g + mask_m + mask_s
+
+
+    if settings.sig2noise_validate:
+        u, v, mask_s2n = sig2noise_val(
+            u, v, s2n,
+            threshold=settings.sig2noise_threshold
+        )
+        # print(f"s2n filter invalidated {sum(mask_s2n.flatten())} vectors")
+        if settings.show_all_plots:
+            plt.quiver(u,v,color='g')
+            plt.show()
+
+        if settings.show_all_plots and sum(mask_s2n.flatten()): # if not all NaN
+            plt.figure()
+            plt.hist(s2n[s2n>0].flatten(),31)
+            plt.show()
+
+        mask += mask_s2n
+
+    return u, v, mask
\ No newline at end of file
